//Source: www.alexandre-pestana.com/volumetric-lights/
//steps for raymarching
#define _STEPS 60
//scattering constant for phase function
#define _SCATTERING 0.75

Texture2D shadowMap : register(t0);
SamplerState clampSamp : register(s0);

cbuffer LightCamera /*Action!*/ : register(b1)
{
	float3 cameraPosition;
	matrix lightProjection;
	float4 lightAmbient;
	float4 lightDiffuse;
	float3 lightPosition;
	matrix lightView;
};

struct InputType
{
	float4 pos : SV_POSITION;
	float2 tex : TEXCOORD0;
	float3 norm : NORMAL;
	float4 shadowMapDepth : TEXCOORD6;
	float4 col : COLOR;
	matrix worldMatrix : TEXCOORD1;
	float4 worldPosition : TEXCOORD5;
};

float4 main(InputType input) : SV_TARGET
{
	if (!input.col.a)
	{
		discard;
	}

	float4 colour = input.col*lightAmbient;

	float3 lightDirection = normalize(input.worldPosition.xyz - lightPosition);

	//calculate vector to travel on
	float3 rayVector = cameraPosition - input.worldPosition.xyz;
	float rayScalar = length(rayVector);
	float3 rayDirection = rayVector / rayScalar;
	float stepLength = rayScalar / _STEPS;
	float3 step = rayDirection * stepLength;
	float3 currentPos = input.worldPosition.xyz;
	float4 fog = float4(0, 0, 0, 0);

	for (int i = 0; i < _STEPS; ++i)
	{
		//get depth and coordinates
		float4 lightPerspective = mul(currentPos, lightView);
		lightPerspective = mul(lightPerspective, lightProjection);
		float lightDepth = lightPerspective.z / lightPerspective.w;
		lightDepth -= 1.f;
		float2 shadowTexCoord = lightPerspective.xy / lightPerspective.z;
		shadowTexCoord.x = (shadowTexCoord.x / 2) + 0.5f;
		shadowTexCoord.y = (-shadowTexCoord.y / 2) + 0.5f;

		//check if it is in range
		if ((saturate(shadowTexCoord.x) == shadowTexCoord.x) && (saturate(shadowTexCoord.y) == shadowTexCoord.y))
		{
			//sample the shadow map
			float shadowMapValue = shadowMap.Sample(clampSamp, shadowTexCoord).r;

			//compare to depth
			if (lightDepth < shadowMapValue)
			{
				//scatter the light as it travels
				float additionalFog = 1 - _SCATTERING*_SCATTERING;
				additionalFog /= (4 * 3.14159f*pow(1 + _SCATTERING*_SCATTERING - (2 * _SCATTERING) * (dot(rayDirection, lightDirection)), 1.5f));
				fog += float4(additionalFog.xxx, 0) * lightDiffuse;

				//only do this for the vertex itself
				if (i == 0)
				{
					//regular lighting
					float intensity = saturate(dot(input.norm, -lightDirection));
					if (intensity > 0)
					{
						colour +=lightDiffuse*intensity;
						colour = saturate(colour);
					}
				}
			}
			else
			{
				float additionalFog = 1 - _SCATTERING*_SCATTERING;
				additionalFog /= (4 * 3.14159f*pow(1 + _SCATTERING*_SCATTERING - (2 * _SCATTERING) * (dot(rayDirection, lightDirection)), 1.5f));
				fog += float4(additionalFog.xxx, 0) * lightAmbient;
			}

		}
		currentPos += step, 0;
		
	}

	//average out the fog
	fog.rgb /= _STEPS;
	colour += fog;

	colour.a = input.col.a;

	return colour;
}