//steps for raymarching
#define STEPS 100
//scattering constant for phase function
#define SCATTERING 0.75

Texture2D text : register(t0);
SamplerState samp : register(s0);
Texture2D shadowMap : register(t1);
SamplerState clampSamp : register(s1);

cbuffer LightCamera /*Action!*/ : register(b2)
{
	float3 cameraPosition;
	float lightSpecPower;
	matrix lightProjection;
	float4 lightAmbient;
	float4 lightDiffuse;
	float3 lightPosition;
	matrix lightView;
	float4 lightSpecColour;
};

struct InputType
{
	float4 position : SV_POSITION;
	float2 tex : TEXCOORD0;
	float3 normal : NORMAL;
	float4 colour : COLOR;
	matrix worldMatrix : TEXCOORD1;
	float4 worldPosition : TEXCOORD5;
};

float4 main(InputType input) : SV_TARGET
{
	//change white in the texture to match the colour for that face
	float4 textureColour = (input.colour + text.Sample(samp, input.tex)) - float4(1.0f, 1.0f, 1.0f, 1.0f);
	float4 colour = lightAmbient * textureColour;
	float4 specular = float4(0, 0, 0, 1);

	//calculate vectors for raymarching
	float3 lightDirection = normalize(input.worldPosition.xyz - lightPosition);

	float3 rayVector = input.worldPosition.xyz - cameraPosition;
	float rayScalar = length(rayVector);
	float3 rayDirection = rayVector/rayScalar;
	float3 step = rayDirection * STEPS;
	
	float4 currentPos = input.worldPosition;
	float4 fog = float4(0, 0, 0, 1);

	for (int i = 0; i < STEPS; ++i)
	{
		//calculate depth and shadow map coordinates every time
		float4 lightPerspective = mul(currentPos, lightView);
		lightPerspective = mul(lightPerspective, lightProjection);
		float lightDepth = lightPerspective.z / lightPerspective.w;
		float2 shadowTexCoord = lightPerspective.xy / lightPerspective.z;
		shadowTexCoord.x = (shadowTexCoord.x / 2) + 0.5f;
		shadowTexCoord.y = (shadowTexCoord.y / 2) + 0.5f;

		//check if this pixel is seen by the light
			if ((saturate(shadowTexCoord.x) == shadowTexCoord.x) && (saturate(shadowTexCoord.y) == shadowTexCoord.y))
			{
				//sample the shadow map
				float shadowMapValue = shadowMap.Sample(clampSamp, shadowTexCoord).r;
				//compare it to the depth
				if (shadowMapValue < lightDepth)
				{
					//henyey-greenstein
					float additionalFog = 1 - SCATTERING*SCATTERING;
					additionalFog /= (4 * 3.14159f*pow(1 + SCATTERING*SCATTERING - (2 * SCATTERING) * (dot(rayDirection, lightDirection)), 1.5f));
					fog += float4(additionalFog.xxx, 0) * lightDiffuse * textureColour;

					//only do specular on the surfave itself
					if (i == 0)
					{
						float intensity = saturate(dot(input.normal, lightDirection));
						if (intensity > 0)
						{
							//blinn-phong
							float3 halfway = normalize(lightDirection + rayVector);
							float spec = pow(max(dot(input.normal, halfway), 0), lightSpecPower);
							specular += saturate(lightSpecColour * spec);
						}

						colour = saturate(colour + specular);
					}
				}
			}
		
		currentPos += float4(step, 0);

	}
	//average out fog
	fog /= STEPS;
	//colour *= textureColour;
	colour = saturate(colour + fog);

	return colour;
}