#include "ShadowMapShader.h"

ShadowMapShader::ShadowMapShader(ID3D11Device* device, HWND hwnd) : BaseShader(device, hwnd)
{
	initShader(L"Puzzle_vs.cso", L"Puzzle_gs.cso", L"Shadow_ps.cso");
}

ShadowMapShader::~ShadowMapShader()
{
	if (matrixBuffer)
	{
		matrixBuffer->Release();
		matrixBuffer = 0;
	}
	BaseShader::~BaseShader();
}

void ShadowMapShader::initShader(WCHAR* vsFilename, WCHAR* psFilename)
{
	loadVertexShader(vsFilename);
	loadPixelShader(psFilename);

	D3D11_BUFFER_DESC matrixBufferDesc;
	D3D11_BUFFER_DESC rotationBufferDesc;
	D3D11_SAMPLER_DESC samplerDesc;

	//matrix buffer description
	matrixBufferDesc.Usage = D3D11_USAGE_DYNAMIC;
	matrixBufferDesc.ByteWidth = sizeof(MatrixBufferType);
	matrixBufferDesc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
	matrixBufferDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
	matrixBufferDesc.MiscFlags = 0;
	matrixBufferDesc.StructureByteStride = 0;

	renderer->CreateBuffer(&matrixBufferDesc, NULL, &matrixBuffer);

	//rotation buffer description
	rotationBufferDesc.Usage = D3D11_USAGE_DYNAMIC;
	rotationBufferDesc.ByteWidth = sizeof(RotationBufferType);
	rotationBufferDesc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
	rotationBufferDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
	rotationBufferDesc.MiscFlags = 0;
	rotationBufferDesc.StructureByteStride = 0;

	renderer->CreateBuffer(&rotationBufferDesc, NULL, &rotationBuffer);

}

void ShadowMapShader::initShader(WCHAR* vs, WCHAR* gs, WCHAR* ps)
{
	//overload initShader, loading the geometry shader then continuing as usual
	loadGeometryShader(gs);
	initShader(vs, ps);
}

void ShadowMapShader::setShaderParameters(ID3D11DeviceContext* deviceContext, const XMMATRIX &world, const XMMATRIX &view, const XMMATRIX &projection, const XMFLOAT4 rotation)
{
	HRESULT result;
	D3D11_MAPPED_SUBRESOURCE mappedResource;
	MatrixBufferType* dataPtr;
	RotationBufferType* rotatPtr;
	unsigned int bufferNumber;
	XMMATRIX tworld, tview, tproj;

	// Transpose the matrices to prepare them for the shader.
	tworld = XMMatrixTranspose(world);
	tview = XMMatrixTranspose(view);
	tproj = XMMatrixTranspose(projection);

	// Lock the constant buffer so it can be written to.
	result = deviceContext->Map(matrixBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);

	// Get a pointer to the data in the constant buffer.
	dataPtr = (MatrixBufferType*)mappedResource.pData;

	// Copy the matrices into the constant buffer.
	dataPtr->world = tworld;
	dataPtr->view = tview;
	dataPtr->projection = tproj;

	// Unlock the constant buffer.
	deviceContext->Unmap(matrixBuffer, 0);

	// Set the position of the constant buffer in the vertex shader.
	bufferNumber = 0;

	// Now set the constant buffer in the vertex shader with the updated values.
	deviceContext->GSSetConstantBuffers(bufferNumber, 1, &matrixBuffer);

	//set up rotation buffer
	result = deviceContext->Map(rotationBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	rotatPtr = (RotationBufferType*)mappedResource.pData;
	rotatPtr->rotation = rotation;
	deviceContext->Unmap(rotationBuffer, 0);
	bufferNumber = 1;
	deviceContext->VSSetConstantBuffers(bufferNumber, 1, &rotationBuffer);
}